
#include <cmath>
#include <zlib.h>
#include "kstring.h"//<-used for buffered output when dumping beagle 0.204
#include "bfgs.h"
#include "analysisFunction.h"


/*
  readerors from an output files
  new version averages from all lines.
 */
void readError(double **errors,const char *fname){
  fprintf(stderr,"will try to read errorestimates from file:%s\n",fname);
  FILE *fp=NULL;
  if(NULL==(fp=fopen(fname,"r"))){
    fprintf(stderr,"Error opening file: %s\n",fname);
    exit(0);
  }
  
  char buf[LENS];
  double res[16];
  for(int i=0;i<16;i++) res[i] = 0;

  int nLines =0;
  while(fgets(buf,LENS,fp)){
    res[0] += atof(strtok(buf," \t\n"));
    for(int i=1;i<16;i++)
      res[i] += atof(strtok(NULL," \t\n"));
    nLines ++;
  }
  for(int j=0;j<16;j++)
    fprintf(stderr,"%f\t",res[j]);
  fprintf(stderr,"\nEstimating errors using nChunks:%d\n",nLines);
  int pos =0;
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      errors[i][j] = res[pos++]/(1.0*nLines);
  fclose(fp);
}


class likeClass{
private:
  double **errors;
  char * errorFname;
  kstring_t bufstr;
  double ****errorProbs;
public:
  //none optional stuff
  FILE *outfile;
  gzFile gzoutfile;
  int doLike;
  int doGlf;
  void run(funkyPars  *pars);
  void printAndClean(funkyPars *pars);  
  void addDefault(funkyPars *pars);
  void openfile(const char *outfiles);
  void getOptions(argStruct arguments);
  likeClass(const char *outfiles,argStruct arguments,int inputtype);
  ~likeClass();
  //other stuff
  int pml_genLikes(funkyPars *pars);
  double **getLikesFullError10Genotypes(int numSites,int nInd,suint **counts,double ****errorProbs,int *keepSites);
  void printLike(funkyPars *pars);
};



void likeClass::getOptions(argStruct arguments){
  //default
  doLike=0;
  doGlf=0;
  errorFname = NULL;
  int doMajorMinor=0;
  int doCounts =0;
  //from command line
  doLike=angsd::getArg("-doLike",doLike,arguments);
  doCounts=angsd::getArg("-doCounts",doCounts,arguments);
  doMajorMinor=angsd::getArg("-doMajorMinor",doMajorMinor,arguments);
  doGlf=angsd::getArg("-doGlf",doGlf,arguments);
  errorFname = angsd::getArg("-errors",errorFname,arguments);
  if(doLike&&(doCounts==0)){
    fprintf(stderr,"Must supply -doCounts\n");
    exit(0);
  }
  if(doGlf==2 &&doMajorMinor==0){
    fprintf(stderr,"For dumping beagle you need to estimate major/minor: -doMajorMinor\n");
    exit(0);
  }
  
  
}

likeClass::likeClass(const char *outfiles,argStruct arguments,int inputtype){
  errorProbs = NULL;
  bufstr.s=NULL; bufstr.l=bufstr.m=0;// <- used for buffered output 
  //default errormatrix
  double errorsDefault[4][4]={    {0,       0.00031, 0.00373, 0.000664},
				  {0.000737    ,    0  ,     0.000576    ,    0.001702},
				  {0.001825 ,       0.000386  ,      0 ,      0.000653},
				  {0.00066, 0.003648 ,       0.000321  ,      0},
				  
  };
  
  getOptions(arguments);
  outfile = NULL;
  gzoutfile = Z_NULL;

  if(doGlf){
    const char* postfix = ".glf";
    const char *beaglepostfix = ".beagle.gz";
    if(doGlf!=2)
      outfile = openFile(outfiles,postfix);
    else
      gzoutfile = openFileGz(outfiles,beaglepostfix,"w6h");
    // fprintf(stderr,"Dumping likelihoods in: %s\n",outfiles);
  }

  if(inputtype==5&&doGlf){
    fprintf(stderr,"Error: cannot output likelihoods (doGlf) when input is beagle\n");
    exit(0);
  }
  if(doGlf&&doLike==0&&inputtype==1){
    fprintf(stderr,"You have do estimate likelihoods (-doLike) for output (-doGlf) \n");
    exit(0);
  }


  if(doLike==0)
    return;

  //inputtype 0) soap 1) samglf 2) samglfClean 3) tglf 4)sim type 5) beagle 7) upppile
  if(inputtype!=0&&inputtype!=7){
    fprintf(stderr,"Error: Likelihoods can only be estimated based on SOAP input and uppile input\n");
    exit(0);
  }
  //allocate and plug in default values
  errors = new double *[4];
  for(int i=0;i<4;i++){
    errors[i] = new double[4];
    for(int j=0;j<4;j++)
      errors[i][j] = errorsDefault[i][j];
  }
  if(errorFname!=NULL)
    readError(errors,errorFname);


  errorProbs = error::generateErrorPointers(errors,3,4);
}


likeClass::~likeClass(){
  error::killGlobalErrorProbs(errorProbs);
  if(doGlf)//filehandle is only open if we want to print GL
    if(outfile!=NULL)
      fclose(outfile);
    else
      gzclose(gzoutfile);

}



void likeClass::printAndClean(funkyPars *pars){

  if(doGlf){
    printLike(pars);
  }
  if(pars->likes!=NULL){
    for(int i=0;i<pars->numSites;i++)
      delete [] pars->likes[i];
    delete [] pars->likes;
  }
  if(pars->post!=NULL){
    for(int i=0;i<pars->numSites;i++)
      delete [] pars->post[i];
    delete [] pars->post;
  }
  if(doLike==0)
    return;
}


void likeClass::run(funkyPars *pars){
  if(doLike==0)
    return;

  pml_genLikes(pars);
  
  //  glob_pars = p;
  // likeFixedMinorError_bfgs(p->errors,p); //est error

}



int likeClass::pml_genLikes(funkyPars *pars){

  pars->likes = getLikesFullError10Genotypes(pars->numSites,pars->nInd,pars->counts,errorProbs,pars->keepSites);
  return 0;
}

double **likeClass::getLikesFullError10Genotypes(int numSites,int nInd,suint **counts,double ****errorProbs,int *keepSites) {

  //only calculate this once
  static float *logfactorial=NULL;
  if(logfactorial==NULL)//dont bother populating if exists.
    logfactorial=error::logfact(LOGMAX); //calculate log factorials

  double **loglikes = new double*[numSites];
  double *logError;


  for(int s=0;s<numSites;s++){
    loglikes[s] = new double [10*nInd];
    if(keepSites[s]==0)
      continue;
    for(int allele1=0;allele1<4;allele1++) {
      for(int allele2=allele1;allele2<4;allele2++){
	int Gindex=angsd::majorminor[allele1][allele2];
	int geno=0;
	int m2=allele2;
	if(allele1!=allele2)
	  geno++;
	else{//total grimt must redo
	m2++;
	 if(m2>3)
	  m2=0;
	}
	logError=errorProbs[geno][allele1][m2];
	for(int i=0;i<nInd;i++){
	  loglikes[s][i*10+Gindex]=logfactorial[counts[s][i*4+0]+counts[s][i*4+1]+counts[s][i*4+2]+counts[s][i*4+3]]; //should be computed before these loops for faster implimentation
	  for(int j=0;j<4;j++)
	    loglikes[s][i*10+Gindex]+=-logfactorial[counts[s][i*4+j]]+counts[s][i*4+j]*logError[j];
	}
      }
    }




    if(0&&s==2){//print infomation about counts and computed likelihood
      for(int i=0;i<nInd;i++){
	fprintf(stderr,"%d\t%d\t%d\t%d\t | %f\t%f\t%f\t",counts[s][i*4+0],counts[s][i*4+1],counts[s][i*4+2],counts[s][i*4+3],loglikes[s][i*3],loglikes[s][i*3+1],loglikes[s][i*3+2]);
	for(int j=0;j<10;j++)
	  fprintf(stderr,"%f\t",loglikes[s][i*10+j]);
	  fprintf(stderr,"\n");
      }


      exit(0);
    }
  }
  return loglikes;
}



void likeClass::printLike(funkyPars *pars) {
  if(pars->likes==NULL)
    return;
  //  pars->opt->printLike=2;
  //fprintf(stderr,"printing likes with option %d\n",pars->opt->doGLF);

  //glffinn format
  if(doGlf==1){
    for(int i=0;i<pars->numSites;i++){
      if(pars->keepSites[i]==0)//never skip, just for debug daemon dragon hack nasty
	continue;
      fwrite(pars->likes[i],sizeof(double),10*pars->nInd,outfile);
    }
    fflush(outfile);
    //    exit(0);
  }
  //beagle format
  else if(doGlf==2){
    int doBuffered = 1;
    
    
    //header of the beagle file
    if(outfile==NULL){
      
      if(pars->chunkNumber==1){
	bufstr.l=0;
	kputs("marker\tallele1\tallele2",&bufstr);
	for(int i=0;i<pars->nInd;i++){
	  kputs("\tInd",&bufstr);
	  kputw(i,&bufstr);
	  kputs("\tInd",&bufstr);
	  kputw(i,&bufstr);
	  kputs("\tInd",&bufstr);
	  kputw(i,&bufstr);
	}
	kputc('\n',&bufstr);
	gzwrite(gzoutfile,bufstr.s,bufstr.l);
      }
    }
    else {
      if(pars->chunkNumber==1){
	fprintf(outfile,"marker\tallele1\tallele2");
	for(int i=0;i<pars->nInd;i++)
	  fprintf(outfile,"\tInd%d\tInd%d\tInd%d",i,i,i);
	fprintf(outfile,"\n");
	
      }

    }
 
    if(doBuffered ==0){
      for(int s=0;s<pars->numSites;s++) {
	if(pars->keepSites[s]==0)
	  continue;
	fprintf(outfile,"%s_%d\t%d\t%d",pars->sites[s].chromo,pars->sites[s].position+1,pars->major[s],pars->minor[s]);
	for(int i=0;i<pars->nInd;i++){
	  double norm=exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]]);
	  fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])/norm);
	  fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])/norm);
	  fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]])/norm);
	}
	//loglike[s][i]);
	fprintf(outfile,"\n");
      }
    }else{
      for(int s=0;s<pars->numSites;s++) {
	bufstr.l = 0; //set tmpbuf beginning to zero
	if(pars->keepSites[s]==0)
	  continue;
	kputs(pars->sites[s].chromo,&bufstr);
	kputc('_',&bufstr);
	kputw(pars->sites[s].position+1,&bufstr);
	kputc('\t',&bufstr);
	kputw(pars->major[s],&bufstr);
	kputc('\t',&bufstr);
	kputw(pars->minor[s],&bufstr);
	//	fprintf(outfile,"%s_%d\t%d\t%d",pars->sites[s].chromo,pars->sites[s].position,pars->major[s],pars->minor[s]);
	for(int i=0;i<pars->nInd;i++) {
	  double norm=exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]]);
	  double val1 = exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])/norm;
	  double val2 = exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])/norm;
	  double val3 = exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]])/norm;
	  ksprintf(&bufstr, "\t%f",val1);
	  ksprintf(&bufstr, "\t%f",val2);
	  ksprintf(&bufstr, "\t%f",val3);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])/norm);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])/norm);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]])/norm);
	}
	//loglike[s][i]);
	kputc('\n',&bufstr);
	if(outfile!=NULL)
	  fprintf(outfile,"%s",bufstr.s);
	else
	  gzwrite(gzoutfile,bufstr.s,bufstr.l);
      }
    }

    ////////
  }
  else{
    //otherwise print textoutput
    //    fprintf(stderr,"[%s]length=%lu\n",__FUNCTION__,10*pars->nInd);
    for(int s=0;s<pars->numSites;s++){
      if(pars->keepSites[s]==0)
	continue;
      fprintf(outfile,"%s\t%d\t",pars->sites[s].chromo,pars->sites[s].position+1);
      for(int i=0;i<10*pars->nInd-1;i++)
	fprintf(outfile,"%f\t",pars->likes[s][i]);
      fprintf(outfile,"%f\n",pars->likes[s][10*pars->nInd-1]);
    }
  }
  fflush(outfile);
}
