#include <cstdio>
#include <cmath>
#include "shared.h"
#include "analysisFunction.h"

#define MINLIKE -1000.0 //this is for setting genotypelikelhoods to missing (EXPLAINED BELOW)

#define SFS ".sfs"
#define SFSPOS ".sfs.pos"

class realSFS{
  int doRealSFS;
  FILE *outfileSFS;
  FILE *outfileSFSPOS;
public:
  //none optional stuff
  FILE *outfile;
  realSFS(const char *outfiles,argStruct arguments,int inputtype);
  //~realSFS();
  void getOptions(argStruct arguments);
  void run(funkyPars  *pars);//printing done here;
  void printAndClean(funkyPars *pars);//printint is done at run...
  
};




void realSFS::getOptions(argStruct arguments){

  
  //from command line
  char *anc=NULL;

  doRealSFS=angsd::getArg("-realSFS",doRealSFS,arguments);
  anc = angsd::getArg("-anc",anc,arguments);
  if(doRealSFS && (anc==NULL)){
    fprintf(stderr,"Must supply -anc for polarizing the spectrum\n");
    exit(0);
  }
}

realSFS::realSFS(const char *outfiles,argStruct arguments,int inputtype){
  //default
  doRealSFS=0;
  getOptions(arguments);
  
  if(doRealSFS==0)
    return;
  
  outfileSFS =  openFile(outfiles,SFS);
  outfileSFSPOS =  openFile(outfiles,SFSPOS);
}


double lbico(double n, double k){
  return lgamma(n+1)-lgamma(k+1)-lgamma(n-k+1);
}


void logrescale(double *ary,int len){
  int maxId = 0;
  //    fprintf(stderr,"maxid:%d maxval:%f\n",maxId,ary[maxId]);
  for(int i=1;i<len;i++){
    //if(posCounter==debug_print)
    //fprintf(stderr,"maxid:%d maxval:%f ary=%f\n",maxId,ary[maxId],ary[i]);
    if(ary[i]>ary[maxId])
      maxId=i;
  }
  double maxVal = ary[maxId];
  //if(posCounter==debug_print)
  // fprintf(stderr,"maxval: %f\n",maxVal);
  for(int i=0;i<len;i++){
    //if(posCounter==debug_print)
    //fprintf(stderr,"%f\t%f\n",ary[i],ary[i]-maxVal);
    ary[i] = ary[i]-maxVal;
  }
}


void algoJoint(double **liks,char *anc,int nsites,int numInds,FILE *sfsfile,int underFlowProtect, int fold,loci *locs,FILE *sfssites,int *keepSites) {
  //  fprintf(stderr,"nsites=%d\tnumInds=%d\n",nsites,numInds);
  fflush(stderr);
  if(anc==NULL||liks==NULL){
    fprintf(stderr,"problems receiving data in [%s] will exit (likes=NULL||ancestral=NULL)\n",__FUNCTION__);
    exit(0);
  }
  double sumMinors[2*numInds+1];  //the sum of the 3 different minors

  for(int it=0; it<nsites; it++) {//loop over sites
    int major_offset = anc[it];
    if(major_offset==4||(keepSites[it]==0))//skip of no ancestral information
      continue;
    //set the resultarray to zeros
    for(int sm=0 ; sm<(2*numInds+1) ; sm++ )
      sumMinors[sm] = 0;
    
    //loop through the 3 different minors
    for(int minor_offset=0;minor_offset<4;minor_offset++) {
      if(minor_offset == major_offset)
	continue;
      double totmax = 0.0;
      //hook for only calculating one minor
      int Aa_offset = angsd::majorminor[minor_offset][major_offset];//0-9
      int AA_offset = angsd::majorminor[minor_offset][minor_offset];//0-9
      int aa_offset = angsd::majorminor[major_offset][major_offset];//0-9
      //      fprintf(stderr,"%d:%d\t%d\t%d\n",major_offset,Aa_offset,AA_offset,aa_offset);
      //part two
      double hj[2*numInds+1];
      for(int index=0;index<(2*numInds+1);index++)
	if(underFlowProtect==0)
	  hj[index]=0;
	else
	  hj[index]=log(0);
      double PAA,PAa,Paa;

      for(int i=0 ; i<numInds ;i++) {
	//	printf("pre scale AA=%f\tAa=%f\taa=%f\n",liks[it][i*3+AA_offset],liks[it][i*3+Aa_offset],liks[it][i*3+aa_offset]);
	double GAA,GAa,Gaa;
#ifdef RESCALE
	if(0){//The rescaling is now done in 'getMajorMinor()'
	double max = liks[it][i*10+0];
	for(int index=1;index<10;index++)
	  if(liks[it][i*10+index]>max)
	    max = liks[it][i*10+index];
	for(int index=0;index<10;index++)
	  liks[it][i*10+index] = liks[it][i*10+index]-max;
	}
#endif
	//printf("post scale AA=%f\tAa=%f\taa=%f\n",liks[it][i*3+AA_offset],liks[it][i*3+Aa_offset],liks[it][i*3+aa_offset]);
	GAA = liks[it][i*10+AA_offset];

	GAa = log(2.0)+liks[it][i*10+Aa_offset];
	Gaa = liks[it][i*10+aa_offset];
	//printf("[GAA] GAA=%f\tGAa=%f\tGaa=%f\n",GAA,GAa,Gaa);
	//do underlfow protection (we are in logspace here) (rasmus style)
	if(1){
	  double mymax;
	  if (Gaa > GAa && Gaa > GAA) mymax = Gaa;
	  else if (GAa > GAA) mymax = GAa;
	  else mymax = GAA;
	  // fprintf(stdout,"mymax[%d]=%f\t",i,mymax);
	  
	  if(mymax<MINLIKE){
	    //	    fprintf(stderr,"\n%f %f %f\n",GAA, GAa,Gaa);
	    Gaa = 0;
	    GAa = 0;
	    GAA = 0;
	    totmax = totmax + mymax;
	  }else{
	    Gaa=Gaa-mymax;
	    GAa=GAa-mymax;
	    GAA=GAA-mymax;
	    totmax = totmax + mymax;
	  }
	//	fprintf(stderr,"totmax=%f\n",totmax);
	//END underlfow protection (we are in logspace here) (rasmus style)
	}
	if(underFlowProtect==0){
	  PAA=exp(GAA);
	  PAa=exp(GAa);
	  Paa=exp(Gaa);
	}else{
	  PAA =(GAA);///(MAA+MAa+Maa);
	  PAa =(GAa);///(MAA+MAa+Maa);
	  Paa =(Gaa);///(MAA+MAa+Maa);
	}

	//check for underflow error, this should only occur once in a blue moon
	if(isnan(Paa)||isnan(PAa)||isnan(Paa)){
	  //fprintf(stderr,"Possible underflow at: \t%d\t%d\n",it->first.chromo,it->first.position);
	  fprintf(stderr,"PAA=%f\tPAa=%f\tPaa=%f\n",PAA,PAa,Paa);
	}
	//fprintf(stdout,"PAA=%f\tPAa=%f\tPaa=%f\n",PAA,PAa,Paa);
	if(i==0){
	  hj[0] =Paa;
	  hj[1] =PAa;
	  hj[2] =PAA;
	}else{
	  
	  for(int j=2*(i+1); j>1;j--){
	    //  print_array(stdout,hj,2*numInds+1,0);
	    //print_array(hj,2*numInds+1);
	    double tmp;
	    if(underFlowProtect==1)
	      tmp =angsd::addProtect3(PAA+hj[j-2],PAa+hj[j-1],Paa+hj[j]);
	    else
	      tmp = PAA*hj[j-2]+PAa*hj[j-1]+Paa*hj[j];
	    
	    if(isnan(tmp)){
	      fprintf(stderr,"is nan:%d\n",j );
	      
	      hj[j] = 0;
	      //	    exit(0);
	      break;
	    }else
	      hj[j]  =tmp;
	  }
	  if(underFlowProtect==1){
	    hj[1] = angsd::addProtect2(Paa+hj[1],PAa+hj[0]);
	    hj[0] = Paa+hj[0];
	  }
	  else{
	    hj[1] = Paa*hj[1] + PAa*hj[0];
	    hj[0] = Paa*hj[0];
	  }
	}
	//ifunderflowprotect then hj is in logspace
	
      }
      //fprintf(stdout,"\nscaledLikes=");
      for(int ii=0;0&&ii<10*numInds;ii++)
	fprintf(stdout,"%f\t",liks[it][ii]);

      //      totmax=0;
      for(int i=0;i<(2*numInds+1);i++)
	if(underFlowProtect==0)
	  sumMinors[i] +=  exp(log(hj[i])-lbico(2*numInds,i)+totmax);
	else
	  sumMinors[i] = exp(angsd::addProtect2(log(sumMinors[i]),hj[i]-lbico(2*numInds,i)+totmax));
    }
    //sumMinors is in normal space, not log
    /*
      we do 3 things.
      1. log scaling everyting
      2. rescaling to the most likely in order to avoid underflows in the optimization
      3. we might do a fold also.
      
     */    

    if(fold){
      fprintf(stderr,"fold is deprecated: will exit\n");
      exit(0);
      int newDim = numInds+1;
      for(int i=0;i<newDim-1;i++)// we shouldn't touch the last element
	sumMinors[i] = log(sumMinors[i] + sumMinors[2*numInds-i]);//THORFINN NEW
      sumMinors[newDim-1] = log(sumMinors[newDim-1])+log(2.0);
      logrescale(sumMinors,newDim);

      fwrite(sumMinors,sizeof(double),newDim,sfsfile);
    }else{
      for(int i=0;i<2*numInds+1;i++)
	sumMinors[i] = log(sumMinors[i]);
      logrescale(sumMinors,2*numInds+1);
      if(isnan(sumMinors[0]))
	fprintf(stderr,"TAYLOR THIS IS AN ERROR: (%s,%d)\n",locs[it].chromo,locs[it].position+1);
      else{
	fprintf(sfssites,"%s\t%d\n",locs[it].chromo,locs[it].position+1);
	fwrite(sumMinors,sizeof(double),2*numInds+1,sfsfile);
      }
    }
    
    //    fprintf(testFP,"\n");
  }
  
}


double gammln(double xx)
{
  double x,y,tmp,ser;
  static double cof[6]={76.18009172947146,-86.50532032941677,
			24.01409824083091,-1.231739572450155,
			0.1208650973866179e-2,-0.5395239384953e-5};
  int j;
  y=x=xx;
  tmp=x+5.5;
  tmp -= (x+0.5)*log(tmp);
  ser=1.000000000190015;
  for (j=0;j<=5;j++) ser += cof[j]/++y;
  return -tmp+log(2.5066282746310005*ser/x);
}


double factln(int n)
{
  static double a[101]; //A static array is automatically initialized to zero.
  if (n < 0) printf("Negative factorial in routine factln: %d \n", n);
  if (n <= 1) return 0.0;
  if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0)); 
  else return gammln(n+1.0);
}


//Func name: bico
//Purpose  : To calculate the nCk ( n combination k)
double bico(int n, int k) {
  return floor(0.5+exp(factln(n)-factln(k)-factln(n-k)));
}

//log version of binomial coef
double lbico(int n, int k){
  //  return log(bico(n,k));
  return factln(n)-factln(k)-factln(n-k);
}


void normalize_array(double *d, int len){
  double s =0;
  for(int i=0;i<len;i++)
    s+=d[i];

  for(int i=0;i<len;i++)
    d[i]=d[i]/s;
}



//Basicly the same as algoBayAll but only looping through the 3 derived given that an ancestral exists;




void print_array(FILE *fp,double *ary,int len,int doLogTransform){
  if(doLogTransform){
    for (int i=0;i<len-1;i++)
      fprintf(fp,"%f\t",log(ary[i]));
    fprintf(fp,"%f\n",log(ary[len-1]));
  }else{
    for (int i=0;i<len-1;i++)
      fprintf(fp,"%f\t",(ary[i]));
    fprintf(fp,"%f\n",(ary[len-1]));
  }
}


  //void algoPopGen(aMap &asso,int numInds,FILE *sfsfile,int underFlowProtect,double *pest) {
void algoPopGen(double **liks,int *anc,int *coverage,loci *locs,int nsites ,int numInds,FILE *sfsfile,int underFlowProtect,double *pest)  {
  fprintf(stderr,"is deprecated will exit\n");
  fflush(stderr);
  exit(0);
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
  //  print_array(stdout,pest,2*numInds+1,0);

  //algorithm goes on by a site on site basis //pretty much the same as 'algo' without the prior phat
  //  numInds=5;

  double scalings[2*numInds-1];
  for(int i=0;i<2*numInds-1;i++){
    double ii=i+1;
    scalings[i] = (ii)*(2*numInds-ii)/bico(2*numInds,2);
  }
  //  fprintf(stdout,"\n");
  //print_array(stdout,scalings,2*numInds-1,0);
  //exit(0);
  double sumMinors[2*numInds+1];  //the sum of the 3 different minors

  for(int it=0; it<nsites; it++) {//loop over sites
  //set the resultarray to zeros
    for(int sm=0 ; sm<(2*numInds+1) ; sm++ )
      sumMinors[sm] = 0;

    int major_offset = anc[it];
    //    int major_offset = p.ancestral;
    //  for(aMap::iterator it=asso.begin(); it!=asso.end(); it++) {//loop over sites
    //datum p = it->second;
    //    fprintf(stdout,"TMP %d %d\n",coverage[it],major_offset);
    if(coverage[it]==0||major_offset==4)
       continue;
    //    fprintf(stderr,"TMP2\n");
    //loop through all nucleotides
      
      //loop through the 3 different minors
      for(int minor_offset=0;minor_offset<4;minor_offset++) {
	if(major_offset==minor_offset)
	  continue;
	double totmax = 0.0;
	//	fprintf(stderr,"comp (%d,%d)\n",major_offset,minor_offset);
	int Aa_offset = angsd::majorminor[minor_offset][major_offset];//0-9
	int AA_offset = angsd::majorminor[minor_offset][minor_offset];//0-9
	int aa_offset = angsd::majorminor[major_offset][major_offset];//0-9
      
	//part two
	double hj[2*numInds+1];
	for(int index=0;index<(2*numInds+1);index++)
	  if(underFlowProtect==0)
	    hj[index]=0;
	  else
	    hj[index]=log(0);
	double PAA,PAa,Paa;
	
	for(int i=0 ; i<numInds ;i++) {
	  //	printf("AA=%f\tAa=%f\taa=%f\n",p.lk[i*3+AA_offset],p.lk[i*3+Aa_offset],p.lk[i*3+aa_offset]);
	  double GAA,GAa,Gaa;
	  GAA = liks[it][i*10+AA_offset];
	  GAa = log(2.0)+liks[it][i*10+Aa_offset];
	  Gaa = liks[it][i*10+aa_offset];
	  	//do underlfow protection (we are in logspace here) (rasmus style)
	if(1){
	  double mymax;
	  if (Gaa > GAa && Gaa > GAA) mymax = Gaa;
	  else if (GAa > GAA) mymax = GAa;
	  else mymax = GAA;
	  // fprintf(stdout,"mymax[%d]=%f\t",i,mymax);
	  
	  if(mymax<MINLIKE){
	    //	    fprintf(stderr,"\n%f %f %f\n",GAA, GAa,Gaa);
	    Gaa = 0;
	    GAa = 0;
	    GAA = 0;
	    totmax = totmax + mymax;
	  }else{
	    Gaa=Gaa-mymax;
	    GAa=GAa-mymax;
	    GAA=GAA-mymax;
	    totmax = totmax + mymax;
	  }
	//	fprintf(stderr,"totmax=%f\n",totmax);
	//END underlfow protection (we are in logspace here) (rasmus style)
	}


	  if(underFlowProtect==0){
	    GAA=exp(GAA);
	    GAa=exp(GAa);
	    Gaa=exp(Gaa);
	  }
	  
	  
	  PAA =(GAA);///(MAA+MAa+Maa);
	  PAa =(GAa);///(MAA+MAa+Maa);
	  Paa =(Gaa);///(MAA+MAa+Maa);


	//check for underflow error, this should only occur once in a blue moon
	  if(isnan(Paa)||isnan(PAa)||isnan(Paa)){
	    //	    fprintf(stderr,"Possible underflow at: \t%d\t%d\n",it->first.chromo,it->first.position);
	    fprintf(stderr,"PAA=%f\tPAa=%f\tPaa=%f\n",PAA,PAa,Paa);
	  }
	  
	  if(i==0){
	    hj[0] =Paa;
	    hj[1] =PAa;
	    hj[2] =PAA;
	  }else{
	    
	    for(int j=2*(i+1); j>1;j--){
	      //print_array(hj,2*numInds+1);
	      double tmp;
	      if(underFlowProtect==1)
		tmp =angsd::addProtect3(PAA+hj[j-2],PAa+hj[j-1],Paa+hj[j]);
	      else
		tmp = PAA*hj[j-2]+PAa*hj[j-1]+Paa*hj[j];
	      
	      if(isnan(tmp)){
		printf("jis nan:%d\n",j );
		//		print_array(stderr,hj,2*numInds+1,0);
		hj[j] = 0;
		//	    exit(0);
		break;
	      }else
		hj[j]  =tmp;
	    }
	    if(underFlowProtect==1){
	      hj[1] = angsd::addProtect2(Paa+hj[1],PAa+hj[0]);
	      hj[0] = Paa+hj[0];
	    }
	    else{
	      hj[1] = Paa*hj[1] + PAa*hj[0];
	      hj[0] = Paa*hj[0];
	    }
	  }

	
	}//after recursion
	//	for(int i=0;i<(2*numInds+1);i++)
	//  hj[i] =  exp(log(hj[i])-log(bico(2*numInds,i)));

	//hj is now in normalspace
	
	//	for(int i=0;i<(2*numInds+1);i++)
	//  sumMinors[i] += hj[i];

	for(int i=0;i<(2*numInds+1);i++)
	  if(underFlowProtect==0)
	    sumMinors[i] +=  exp(log(hj[i])-lbico(2*numInds,i)+totmax);
	  else
	    sumMinors[i] = exp(angsd::addProtect2(log(sumMinors[i]),hj[i]-lbico(2*numInds,i)+totmax));


      }
      //      print_array(stdout,sumMinors,2*numInds+1,0);

    normalize_array(sumMinors,2*numInds+1);
    //print_array(stdout,sumMinors,2*numInds+1,0);
    //first update the noprior SFS
//
 //   for(int i=0;i<2*numInds+1;i++)
 //     raw[i] = sumMinors[i];
//


    //formula 1
    double *workarray = NULL;
    double formula1[2*numInds+1];
    if(pest!=NULL){
      for(int i=0;i<2*numInds+1;i++)
	formula1[i] = sumMinors[i]*pest[i];
      normalize_array(formula1,2*numInds+1);
      workarray = formula1;
    }else
      workarray = sumMinors;
    if(1)
      print_array(sfsfile,workarray,2*numInds+1,0);
    else
      fwrite(workarray,sizeof(double),2*numInds+1,sfsfile);
    
    if(0){//debug code
      double tt[] = {0.070, 0.059, 0.082, 0.043, 0.024, 0.087, 0.080, 0.112, 0.073, 0.194, 0.176};
      workarray = tt;
      numInds=5;
      fprintf(stdout,"\n");
      //    print_array(stdout,workarray,2*numInds+1,0);
      print_array(stdout,scalings,2*numInds-1,0);
    }
    double seq =  1-workarray[0]-workarray[2*numInds];
    double pairwise=0;
    for(size_t i=1;i<2*numInds;i++){//skip last element{
      //fprintf(stderr,"%lu: %f %f=%f\n",i,workarray[i],scalings[i-1],workarray[i] *scalings[i-1]);
      pairwise += workarray[i] *scalings[i-1];
    }
    fprintf(stdout,"%s\t%d\t%f\t%f\n",locs[it].chromo,locs[it].position,log(seq),log(pairwise));
    //fprintf(stdout,"%d\t%f\t%f\n",it->first.position,log(seq),log(pairwise));
    
    //print_array(stdout,formula1,2*numInds+1,0);
    //exit(0);
    continue;
  }  

}
//printing done here;
void realSFS::run(funkyPars  *p){
  if(doRealSFS==0)
    return;
  else if(doRealSFS==1){
    algoJoint(p->likes,p->anc,p->numSites,p->nInd,outfileSFS,0,0,p->sites,outfileSFSPOS,p->keepSites);
  }else{
    fprintf(stderr,"unsupported realSFS arg=%d\n",doRealSFS);
  }


}

void realSFS::printAndClean(funkyPars *p){


}
