#include<fstream> //ifstream requires fstream
#include <list>
#include <vector>
#include <cstring>
#include "shared.h"
#include "parseArgs.h"
#include "analysisFunction.h"

//different MAF estimators
#define MAF_KNOWN_BFGS 0x1
#define MAF_KNOWN_EM 0x2
#define MAF_UNKNOWN_BFGS 0x4
#define MAF_UNKNOWN_EM 0x8
#define EM_START 0.001

//stuff related to genotypecalling
#define GENO_MAJOR_MINOR 0x1 // write the major and the minor
#define GENO_MAF_PRIOR 0x2 //print the called genotype
#define GENO_TAKE_MAX 0x4 //just pick the maxim likelihood
#define GENO_WRITE_POST 0x8 //write the post of the called genotype
#define GENO_ALL_POST 0x10 // write the post
#define GENO_WRITE_DEPTH 0x20 //write the depth
#define GENO_FOR_COVAR 0x40 //binary dump of the posteriors.



void parseArgStruct(argStruct arguments){
  fflush(stderr);
  for(int i=1;i<arguments.argc;i++){
    if(arguments.usedArgs[i]==0){
      fprintf(stderr,"%d argument \t%s is unknown will exit\n",i,arguments.argv[i]);
      exit(0);
    }
  }
  fflush(stderr);
  // exit(0);
}


double *readArray(char *fname){
  FILE *fp=getFILE(fname,"r");
  char buf[fsize(fname)];
  if(fsize(fname)!=fread(buf,fsize(fname),1,fp))
    fprintf(stderr,"Probmeasdfasdf\n");
  std::vector<double> vec;
  vec.push_back(atof(strtok(buf,"\t \n")));
  while(1){
    char *tok = strtok(NULL,"\t \n");
    if(tok ==NULL)
      break;
    vec.push_back(atof(tok));
  }
  fprintf(stderr,"\t-> nVals in: %lu in file:%s\n",vec.size(),fname);
  double *ret = new double[vec.size()];
  for(size_t i=0;i<vec.size();i++)
    ret[i] = vec[i];
  return ret;
}


void getNind(const char *name,argStruct arguments){
 if(!fexists(name)){
    fprintf(stderr,"\t-> Problems opening file: %s\n",name);
    exit(0);
  }

 const char* delims = " \t";
 std::ifstream pFile(name,std::ios::in);
 char buffer[LENS];
 int nInd=0;
 while(!pFile.eof()){
   pFile.getline(buffer,LENS);
   nInd++;
 }
 fprintf(stdout,"Nind=%d in %s\n",nInd,name);
 arguments.nInd=nInd;
 arguments.inputFilename=strdup(name);
}



Matrix<double> getMatrix(const char *name,int doBinary){
  if(!fexists(name)){
    fprintf(stderr,"\t-> Problems opening file: %s\n",name);
    exit(0);
  }
  const char* delims = " \t";
  std::ifstream pFile(name,std::ios::in);
  
  char buffer[LENS];
  std::list<double *> rows;
  int ncols =0;
  while(!pFile.eof()){
    pFile.getline(buffer,LENS);
    if(strlen(buffer)==0)
      continue;
    char *tok = strtok(buffer,delims);
    std::list<double> items;
    while(tok!=NULL){
      if(doBinary)
	items.push_back(atoi(tok));
      else
	items.push_back(atof(tok));
      tok = strtok(NULL,delims);
    }
    //fprintf(stderr,"[%s] ncols:%lu\n",__FUNCTION__,items.size());
    ncols = items.size();
    double *drows = new double[items.size()];
    int i=0;
    for(std::list<double>::iterator it=items.begin();it!=items.end();it++)
      drows[i++]  = *it;
    rows.push_back(drows);
    
  }
  fprintf(stderr,"%s nrows:%lu\n",__FUNCTION__,rows.size());
  double **data = new double*[rows.size()];
  int i=0;
  for(std::list<double*>::iterator it=rows.begin();it!=rows.end();it++)
    data[i++]  = *it;
  
  Matrix<double> retMat;
  retMat.matrix=data;
  retMat.x = rows.size();
  retMat.y = ncols;
  return retMat;

}



progArgs *allocProgArgs(){
  progArgs* ret = new progArgs;

  ret->doLike=0;
  ret->doPost=0;

  ret->sfsEst = NULL;
  ret->flag = 0;
  //sort data stuff
  ret->minMaf=0;
  ret->minInd=0;
  
  //soap stuff;
  ret->qs =0;
  ret->strand = -1;
  ret->doError =0;
  ret->maxHits=1;
  ret->cutOff=0.01;
  ret->downsample=0;
  ret->minDepth = 200;
  ret->maxDepth= 50000;
  ret->eps = 0.001;


  ret->sim1 = NULL;
  ret->samglf = NULL;
  ret->beagle = NULL;
  ret->samglfclean = NULL;
  ret->soap = NULL;
  ret->tglf = NULL;
  ret->outfiles = NULL;
  ret->faifile = NULL;
  ret->target = NULL;
  ret->refName = NULL;
  ret->ancName =NULL;
  ret->posfile = NULL;
  ret->realSFS = 0;
  ret->doGeno = 0;
  ret->doMaf =0;//0=dont ;1=
  ret->doGlf = 0;
  ret->doAsso =0;
  ret->assoCutoff = 0.001;
  ret->nLines = 1000;
  ret->ymat.x=0;
  ret->ymat.y=0;
  ret->ymat.matrix=NULL;
  
  ret->covmat.x=0;
  ret->covmat.y=0;
  ret->covmat.matrix=NULL;
  ret->sitePerm=1;  
  ret->minHigh =10;
  ret->minCount=10;
  ret->emIter=100;
  ret->getCovar=0;
  ret->adjust=-1;
  ret->chunkSize = 10000;//numbers of sites above cutoff that should be used for further anylsis
  
  ret->errors = NULL;
  ret->inputtype =-1;
  ret->mpileup =0;
  ret->uppile = 0;
  double errors[4][4]={    {0,       0.00031, 0.00373, 0.000664},
			   {0.000737    ,    0  ,     0.000576    ,    0.001702},
			   {0.001825 ,       0.000386  ,      0 ,      0.000653},
			   {0.00066, 0.003648 ,       0.000321  ,      0},
			   
  };
  
  double **tmps = new double *[4];
  for(int i=0;i<4;i++)
    tmps[i] = new double[4];
  
  for(int i=0; i<4; i++)
    for(int j=0; j<4; j++) 
      tmps[i][j] = errors[i][j];
  ret->errors = tmps;
  
  
  
  ret->yfile = NULL;
  ret->covfile = NULL;
  ret->lStart =NULL;
  ret->lStop  =NULL;

  ret->nThreads = 1;
  ret->nInd = -1;

  ret->doSNP = 0;
  return ret;
}


std::vector<char*> getFilenames(char * name){
  if(!fexists(name)){
    fprintf(stderr,"\t-> Problems opening file: %s\n",name);
    exit(0);
  }
  const char* delims = " \t";
  std::vector<char*> ret;
  std::ifstream pFile(name,std::ios::in);

  char buffer[LENS];
  while(!pFile.eof()){
    pFile.getline(buffer,LENS);
    char *tok = strtok(buffer,delims);
    while(tok!=NULL){
      if(tok[0]!='#')
	ret.push_back(strdup(buffer));
      tok = strtok(NULL,delims);
    }
  }
  return ret;
}

std::vector<char*> getSingleFilenames(char * name){
  if(!fexists(name)){
    fprintf(stderr,"\t-> Problems opening file: %s\n",name);
    exit(0);
  }
  //  const char* delims = " \t";
  std::vector<char*> ret;
  ret.push_back(strdup(name));

  return ret;
}



void readErrorOld(double **errors,const char *fname){
  FILE *fp;
  if(NULL==(fp=fopen(fname,"r"))){
    fprintf(stderr,"Error opening file: %s\n",fname);
    exit(0);
  }
  char buf[LENS];
  if(fgets(buf,LENS,fp)==NULL){
    fprintf(stderr,"file: %s looks empty\n",fname);
    exit(0);
  }
  errors[0][0] = atof(strtok(buf," \t"));
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      if(i==0&&j==0)
	continue;
      else
	errors[i][j] = atof(strtok(NULL," \t")); 
  fclose(fp);
}

/*
  new version averages from all lines.
 */
void readError_local(double **errors,const char *fname){
  FILE *fp=NULL;
  if(NULL==(fp=fopen(fname,"r"))){
    fprintf(stderr,"Error opening file: %s\n",fname);
    exit(0);
  }
  
  char buf[LENS];
  double res[16];
  for(int i=0;i<16;i++) res[i] = 0;

  int nLines =0;
  while(fgets(buf,LENS,fp)){
    res[0] += atof(strtok(buf," \t\n"));
    for(int i=1;i<16;i++)
      res[i] += atof(strtok(NULL," \t\n"));
    nLines ++;
  }
  for(int j=0;j<16;j++)
    fprintf(stderr,"%f\t",res[j]);
  fprintf(stderr,"\nEstimating errors using nChunks:%d\n",nLines);
  int pos =0;
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      errors[i][j] = res[pos++]/(1.0*nLines);
  fclose(fp);
}


void printProgArgs(progArgs *a,FILE *fp){
  fprintf(fp,"\n\t[%s]------------------\n",__FUNCTION__);
  fprintf(fp,"\t input files:\n");
  fprintf(fp,"\t\tsamglf=%s\tsamglfclean=%s\tsoap=%s\ttglf=%s\tposfile=%s\tsim1=%s\n",a->samglf,a->samglfclean,a->soap,a->tglf,a->posfile,a->sim1);

  fprintf(fp,"\toutput files (prefix)\n");
  fprintf(fp,"\t\t outfiles=%s\n",a->outfiles);

  fprintf(fp,"\n\tcofiles=%s\tyfile=%s\n",a->covfile,a->yfile);

  fprintf(fp,"\t branch variables\n");
  fprintf(fp,"\t doMaf=%d\tdoGlf=%d\tdoAsso=%d\n",a->doMaf,a->doGlf,a->doAsso);
  fprintf(fp,"\t nLines=%d\tassoCutoff=%f\n",a->nLines,a->assoCutoff);
  fprintf(fp,"\t sitePerm=%d\tminHigh=%d\tminCount=%d\temIter=%d\n",a->sitePerm,a->minHigh,a->minCount,a->emIter);
  fprintf(fp,"\t getCover=%d\tadjust=%d\tchunkSize=%d\n",a->getCovar,a->adjust,a->chunkSize);
  //fprintf(fp,"\t errors=%p\tultrastart=(%d,%d),\tultrastop=(%d,%d)\n",a->errors,a->ultraStart.chromo,a->ultraStart.position,a->ultraStop.chromo,a->ultraStop.position);
  fprintf(fp,"\t errors=%p\n",a->errors);
  //  fprintf(fp,"\t faiIndex.size()=%lu\tlStart=%s\tlStop=%s\n",a->faiIndex.size(),a->lStart,a->lStop);
  fprintf(fp,"\tlStart=%s\tlStop=%s\n",a->lStart,a->lStop);
  fprintf(fp,"\tsoapStuff \n");
  fprintf(fp,"\tqs=%d\tstrand=%d\tdoError=%d\n",a->qs,a->strand,a->doError);
  fprintf(fp,"\t\tmaxHits=%d\tcutOff=%f\tdownSample=%d\tminDepth=%d\tmaxDepth=%d\n",a->maxHits,a->cutOff,a->downsample,a->minDepth,a->maxDepth);
  fprintf(fp,"\teps=%f\n",a->eps);
  fprintf(fp,"\tnThreads=%d\n", a->nThreads);
  fprintf(fp,"\ttarget file=%s\n",a->target);
  fprintf(fp,"\t[%s]------------------\n",__FUNCTION__);
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      fprintf(fp,"%f\t",a->errors[i][j]);
  fprintf(fp,"\n");
  fprintf(fp,"\t[%s]------------------\n",__FUNCTION__);
  fprintf(fp,"ref=%s\tanc=%s\n",a->refName,a->ancName);
  fprintf(fp,"opt->minInd=%d\t-minMaf=%f\t-minLRT\n",a->minInd,a->minMaf,a->minLRT);
  fprintf(fp,"doSNP=%d\n",a->doSNP);
  fprintf(fp,"-realSFS = %d\n",a->realSFS);
  fprintf(fp,"-beagle = %s\n",a->beagle);
}

//program to check if files exists and are nonempty
void isOk(char *fname){
  if(fname!=NULL){
    if(!fexists(fname)){
      fprintf(stderr,"\nfilename: %s looks nonexistant,will exit\n",fname);
      exit(0);
    }if(0==fsize(fname)){
      fprintf(stderr,"\nfilename: %s looks empty, will exit\n",fname);
      exit(0); 
      
    }
  }
}



progArgs * parseArgs(int argc,char **argv,argStruct arguments) { 


  progArgs *args = allocProgArgs();

  //these are user supplied parameters
  
  int argPos = 1;
  while(argPos <argc){ 
    if (strcmp(argv[argPos],"-samglf")==0){
      arguments.inputtype=1;
      getNind(argv[argPos+1],arguments);
      args->samglf = (argv[argPos+1]);
      args->filenames = getFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    } else if (strcmp(argv[argPos],"-samglfclean")==0){
      getNind(argv[argPos+1],arguments);
      arguments.inputtype=2;
      args->samglfclean = (argv[argPos+1]);
      args->filenames = getFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    } else if (strcmp(argv[argPos],"-sim1")==0){
      arguments.inputtype=4;
      arguments.nInd=angsd::getArg("-nInd",arguments.nInd,arguments);
      std::vector<char *> tmp;
      args->filenames = tmp;
      args->sim1 = (argv[argPos+1]);
      args->filenames.push_back(strdup(args->sim1));
      //args->filenames = getFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if (strcmp(argv[argPos],"-beagle")==0){
      arguments.inputtype=5;
      args->beagle = (argv[argPos+1]);
      args->filenames = getSingleFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if (strcmp(argv[argPos],"-soap")==0){
      getNind(argv[argPos+1],arguments);
      arguments.inputtype=0;
      args->soap = (argv[argPos+1]);
      args->filenames = getFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    } else if (strcmp(argv[argPos],"-tglf")==0){
      arguments.inputtype=3;
      getNind(argv[argPos+1],arguments);
      args->tglf = (argv[argPos+1]);
      args->filenames = getFilenames(argv[argPos+1]); 
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    } else if (strcmp(argv[argPos],"-posfile")==0){
      args->posfile = (argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
     else if(strcmp(argv[argPos],"-outfiles")==0){
      args->outfiles = argv[argPos+1];
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-realSFS")==0)
      args->realSFS = atoi(argv[argPos+1]);
    else if(strcmp(argv[argPos],"-start")==0){
      args->lStart=(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
 
    }
    else if(strcmp(argv[argPos],"-stop")==0){
      args->lStop=(argv[argPos+1]);
     arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-chunkSize")==0){
      args->chunkSize = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-fai")==0){
      args->faifile = argv[argPos+1];
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-nLines")==0){
      args->nLines = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-qs")==0){
      args->qs = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
    }
    else if(strcmp(argv[argPos],"-strand")==0){
      args->strand = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
      }
    else if(strcmp(argv[argPos],"-maxHits")==0){
      args->maxHits = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-cutOff")==0){
      args->cutOff = atof(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-downsample")==0){
      args->downsample = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-minDepth")==0){
      args->minDepth = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-ref")==0){
      args->refName = argv[argPos+1];
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-anc")==0){
      args->ancName = argv[argPos+1];
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-maxDepth")==0){
      args->maxDepth = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-eps")==0){
      args->eps = atof(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-sfsEst")==0){
      args->sfsEst = readArray(argv[argPos+1]);
       arguments.usedArgs[argPos]=1;
       arguments.usedArgs[argPos+1]=1;
     }  
    else if(strcmp(argv[argPos],"-nThreads")==0){
      args->nThreads = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
    else if(strcmp(argv[argPos],"-target")==0){
      args->target = (argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
     }
      else if(strcmp(argv[argPos],"-nInd")==0){
      fprintf(stderr,"settting nind\n");
      args->nInd = atoi(argv[argPos+1]);
      arguments.usedArgs[argPos]=1;
      arguments.usedArgs[argPos+1]=1;
      } else if(strcmp(argv[argPos],"mpileup")==0){ //deprecated
      fprintf(stderr,"Will use samtools as filereader\n");
      args->mpileup = argPos;
      for(int aa=argPos;aa<argc;aa++){
	arguments.usedArgs[aa]=1;
	arguments.usedArgs[aa]=1;
      }
      break;
    } else if(strcmp(argv[argPos],"-bam")==0){
	getNind(argv[argPos+1],arguments);
	arguments.inputtype=7;
	//	arguments.inputtype=1;
	fprintf(stderr,"Will use bambi as filereader\n");
	args->uppile = argPos-1;
	arguments.usedArgs[argPos]=1;
	arguments.usedArgs[argPos+1]=1;
	break;
      }
    argPos+=2;
  }
  if(!args->posfile != !args->tglf){
    fprintf(stderr,"Must supply posfile and tglf together\n");
    exit(0);
  }

  if(args->outfiles ==NULL){
    fprintf(stderr,"\t-> Must supply -outfiles\n");
    exit(0);
  }

  if(args->soap!=NULL)
    args->inputtype = 0;
  else if(args->samglf!=NULL)
    args->inputtype=1;
  else if(args->samglfclean!=NULL)
    args->inputtype=2;
  else if(args->beagle!=NULL)
    args->inputtype=5;
  else if(args->tglf!=NULL)
    args->inputtype=3;
  else if(args->sim1!=NULL){
    if(args->nInd==-1){
      fprintf(stderr,"\t-> Must set number of individuals when using simfiles");
      exit(0);
    }

    args->inputtype=4;
  }else if(args->mpileup!=0)
    args->inputtype=6;
  else if(args->uppile!=0){

    args->inputtype=7;
    fprintf(stderr,"inputtype is uppile=%d\n",args->inputtype);
  }
  if(args->inputtype<6){
    if(args->filenames.size()==0 ){
      fprintf(stderr,"\t-> Need to supply -(inputfiles) ADD INFO\n");
      exit(0);
      return 0;
    }
  }
  
  //validate the files, before doing anything else
  if(args->mpileup==0)
    for(size_t i=0;i<args->filenames.size();i++){
      isOk(args->filenames[i]);
    }
  isOk(args->covfile);
  isOk(args->yfile);
  isOk(args->faifile);

  if(args->nThreads>1)
    if(0&&(args->doMaf &MAF_KNOWN_BFGS||args->doMaf &MAF_UNKNOWN_BFGS)){
      fprintf(stderr,"\t-> bfgs algorithm is not implemented for multithreading, use diffent MAF estimator\n");
      exit(0);
    }
  
  
  if(args->target!=NULL){
    fprintf(stderr,"\n\t WARNING!!!!! You have choosen a target region\n");
    fprintf(stderr,"\n\t WARNING!!!!! You might want to change the following parameters\n");
    fprintf(stderr,"\n\t WARNING!!!!! -cutOff -minDepth -maxDepth\n");
    fprintf(stderr,"\n\t WARNING!!!!! I'm giving you 10 seconds to think about this before continuing\n");
    fprintf(stderr,"\t ");
    for(int i=0;1&i<10;i++){
      fprintf(stderr,"%d ",i+1);
      sleep(1);
    }
    fprintf(stderr,"\n");
  }
  if(args->sim1==NULL && args->nInd!=-1){
    if(args->filenames.size()<args->nInd)
      fprintf(stderr,"\t-> Number of samples is smaller than subset requested\n");
    else{
      fprintf(stderr,"\t-> Will remove tail of filename list\n");
      args->filenames.erase(args->filenames.begin()+args->nInd,args->filenames.end());
      fprintf(stderr,"\t->  filename list now contains only: %lu\n",args->filenames.size());
    }
    for(size_t ii=0;1&&ii<args->filenames.size();ii++)
      fprintf(stderr,"%lu->%s\n",ii,args->filenames[ii]);
    fprintf(stderr,"\n");
  }
  if(args->realSFS && args->nThreads!=1){
    fprintf(stderr,"\t-> realSFS only implemented for single thread: will set to 1\n");
    args->nThreads = 1;
  }
  fprintf(stderr,"flag=%d\tdoGeno=%d\n",args->flag,args->doGeno);


  if(args->doGeno&GENO_WRITE_POST||args->doGeno&GENO_FOR_COVAR || args->doGeno&GENO_ALL_POST ){
    fprintf(stderr,"\n\t-> For dumping post in genotypecalling you need the MAF-model\n\t Setting GENO_MAF_PRIOR\n\n");
    args->doGeno =     args->doGeno|GENO_MAF_PRIOR;
  }

  if((args->doGeno&GENO_MAF_PRIOR ) &&(args->doMaf==0)){
    fprintf(stderr,"\n\t-> You need a maf estimator for this genotypecaller\n\twill use EM maf\n\n");
    args->doMaf = args->doMaf|MAF_KNOWN_EM;

  }


  return args;

}

void killProgArgs(progArgs *p){
  for(int i=0;i<p->filenames.size();i++)
    free(p->filenames[i]);

  for(int i=0;i<4;i++)
    delete [] p->errors[i];
  delete [] p->errors;
  delete p;
}
