int modSites = 0;
#include <cstdio>
#include <vector>
#include <cstring>
#include <list>
#include <sys/stat.h>
#include <pthread.h>

/*
  order of includes is maybe important
*/

#include "covarClass.cpp"
#include "majorminor.cpp"
#include "estMaf.cpp"
//#include "analysisEstError.cpp"
#include "analysisEstError.h"
#include "analysisEstLikes.cpp"
#include "analysisAsso.cpp"
#include "analysisHWE.cpp"
//tsk below
#include "analysisCallGenotypes.cpp"
#include "angsd_realSFS.cpp"
#include "getFasta.cpp"//for reading fasta; ancestral and refernce
#include "gen_likes.h"
#include "analysisCount.cpp" //generate counts from reads

#include "shared.h"
#include "printRes.h"

/*
  below not arrays but single objects
 */
likeClass *likesObj = NULL;//analysisEstLikes.cpp
covar *covarObj = NULL; //covarClass.cpp
majorminor *majorminorObj = NULL; //majorminor.cpp
frequency *freqObj = NULL;//estMaf.cpp
error *errorObj = NULL;//analysisEstError.c/cpp
asso *assoObj = NULL;//analysisAsso.cpp
hwe *hweObj = NULL;//analysisHWE.cpp
//tsk below
callGenotypes *genotypesObj = NULL;//analysisCallGenotypes.cpp
realSFS *realSFSobj = NULL;
getFasta *fastaObj = NULL;
gen_likes *gen_likesObj = NULL;
countCls *cntObj = NULL;

int maxThreads;//this value is given from the commandline -nThreads
pthread_t thread1;
pthread_cond_t cvMaxThread;
pthread_mutex_t counterMut;
int curRunning =0;
int chunkNumber =1;
int n_phenotypes;
printRes printer;


//void init(const char *outfiles,Matrix<double> matY,Matrix<double> matCov,int doMaf,int doGLF,int doAsso,int doError,int inputtype,int maximalThread,char *ref,char*anc,int doSNP,int doSFS,int dumpCounts,int doGeno,argStruct arguments){
void init(const char *outfiles,int maximalThread,int inputtype,argStruct arguments){
  //  dumpType=dumpCounts;
  maxThreads=maximalThread;
  pthread_mutex_init(&counterMut, NULL);
  pthread_cond_init (&cvMaxThread, NULL);
  //  initilize function objects - classy  //
  likesObj = new likeClass(outfiles,arguments,inputtype);
  freqObj = new frequency(outfiles,arguments,inputtype);
  assoObj = new asso(outfiles,arguments,inputtype);
  hweObj = new hwe(outfiles,arguments,inputtype);
  errorObj = new error(outfiles,arguments,inputtype);
  majorminorObj = new majorminor(outfiles,arguments,inputtype);

  covarObj = new covar(outfiles);
  //tsk below
  genotypesObj = new callGenotypes(outfiles,arguments,inputtype);
  realSFSobj = new realSFS(outfiles,arguments,inputtype);
  fastaObj = new getFasta(arguments);
  //  gen_likesObj is initialize by default in mUpPile, nasty hack dragon daemon
  gen_likesObj = new gen_likes(arguments);
  //
  cntObj = new countCls(outfiles,arguments,inputtype);
}
extern pthread_mutex_t mUpPile_mutex;
void destroy(){
  fprintf(stderr,"[%s:%s()] calling destroy\n",__FILE__,__FUNCTION__);
  while(1){
    pthread_mutex_lock(&counterMut);
    if(curRunning==0)
      break;
    fprintf(stderr,"Waiting for nThreads:%d\n",curRunning);
    pthread_mutex_unlock(&counterMut);
    sleep(1);
  }
  pthread_mutex_unlock(&counterMut);



  //  fprintf(stderr,"[%s] calling destructors from \n",__FUNCTION__);
  //fflush(stderr);

  // close outfiles and other cleaning - classy //
  delete likesObj;
  delete hweObj;
  delete covarObj;
  delete freqObj;
  delete assoObj;
  delete majorminorObj;
  delete errorObj;
  delete genotypesObj;
  delete realSFSobj;
  delete gen_likesObj;
  delete fastaObj;
  delete cntObj;
}

void interface(funkyPars *p){
  //  fprintf(stderr,"[%s] chunknumber=%d\n",__FUNCTION__,p->chunkNumber);
  fflush(stderr);
  if(p->killSig==1){
    fprintf(stderr,"chunkNumber=%d is the killSig\n",p->chunkNumber);
    //    return ;
  }

  if(p->killSig==0)
    main_analysis(p);//dont do analysis on the killsig chunk

  //  if(p->killSig==0)
  printer.prop(p);
}

void *slave(void *ptr){
  funkyPars *p = (funkyPars *) ptr;
  interface(p);

  //this might moved to the interface to allow more chunks with data (no waiting for print)
  pthread_mutex_lock( &counterMut );
  curRunning--;
  pthread_cond_signal(&cvMaxThread);
  pthread_mutex_unlock( &counterMut );

}


void *slave2(void *ptr){
  funkyPars *p = (funkyPars *) ptr;
  if(p->killSig==0)
    main_analysis(p);
  printFunky(p);
  pthread_mutex_unlock( &counterMut );
}

void master2(funkyPars *p){
  //  fprintf(stderr,"[%s] Number of threads running=%d\n",__FUNCTION__,curRunning);
  pthread_mutex_lock( &counterMut );
  int iret1 = pthread_create( &thread1, NULL, slave2, (void*) p);
  pthread_detach(thread1);
}



void master(funkyPars *p){
  //  fprintf(stderr,"[%s] Number of threads running=%d\n",__FUNCTION__,curRunning);
  pthread_mutex_lock( &counterMut );
  curRunning++;
  if(curRunning==maxThreads){
    //fprintf(stderr,"We are running max number of threads will wait for finishing threads: %d\n",curRunning);
    pthread_cond_wait(&cvMaxThread, &counterMut);
    //fprintf(stderr,"Done waiting for finishing threads:%d\n",curRunning);
  }pthread_mutex_unlock( &counterMut );
  int iret1 = pthread_create( &thread1, NULL, slave, (void*) p);
  pthread_detach(thread1);
}

/*
  This is the function that determines whether or not to start a new thread
*/
void selector(funkyPars *p){
  if(p==NULL){
    p = allocFunkyPars();
    p->killSig = 1;
  }
  p->chunkNumber = chunkNumber++;
  
  if(maxThreads==1)
    interface(p);
  else if(maxThreads==2)//simple serialization, allow one chunk to be computated
    master2(p);
  else
    master(p);
}

////////////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////////


loci *populateLoci(const chunkyT *chk){
  loci *ret = new loci[chk->nSites];
  for(int i=0;i<chk->nSites;i++){
    ret[i].chromo = strdup(chk->refName);
    ret[i].position = chk->nd[i][0].refPos;
  }

  return ret;
}




void cleanUptNodeArray(tNode *row,int nSamples){
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
    for(int i=0;i< nSamples;i++) {
      if(row[i].l2!=0)
	//	fprintf(stderr,"l3\t%d\n",chk->nd[s][i].l2);
	for(int j=0;j<row[i].l2;j++){
	//	fprintf(stderr,"removing indels\n");
	dalloc_node(row[i].insert[j]);
      }
      free(row[i].insert);
      dalloc_node(row[i]);
    }
    delete [] row;

  
}


void modify_keepsites(funkyPars *p){
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
  static int *hapsites=NULL;//DRAGON DAEMNon leak
  static int curPos = 0;
  static int nSites =0;
  if(hapsites==NULL){
    static const char *fname = "/home/thorfinn/chr20.cleaned.txt";
    FILE *fp=fopen(fname,"r");
    fprintf(stderr,"fsize of: %s is:%lu\n",fname,fsize(fname));
    char *buf = new char[fsize(fname)];
    if(fsize(fname)!=fread(buf,1,fsize(fname),fp))
      fprintf(stderr,"problems reading full file\n");
    int nLines =0;
    for(size_t i=0;i<fsize(fname);i++)
      if(buf[i]=='\n')
	nLines++;
    //    fprintf(stderr,"nLines=%d\n",nLines);
    hapsites = new int[nLines -1];
    nSites = nLines -1;
    strtok(buf,"\n");
    for(int i=0;i<nSites;i++){
      hapsites[i] = atoi(strtok(NULL," \n\t"))-1;//offset with one
      strtok(NULL,"\n");
    }
    fprintf(stderr,"nSites: to be used as filter:%d\n",nSites);
    delete [] buf;
    fclose(fp);
  }
  for(int i=0;i<p->numSites;i++){
    p->keepSites[i] =0;
    if(curPos==nSites)
      continue;
    else if(p->sites[i].position==hapsites[curPos])
      p->keepSites[i] =1;
    else if(p->sites[i].position>=hapsites[curPos])
      curPos++;
  }
}




chunkyT *splitFucker3000T(nodePoolT *dn,int nFiles);//from mUppile.cpp;

int main_analysis(funkyPars *p) {
  

  //first step is to make a chunk of data from the sample "uppiles"
  if(p->for_callback!=NULL) {
    fcb *f = p->for_callback;
    //    fprintf(stderr,"picking up fcb in main_analysis\n");
    chunkyT *chk = splitFucker3000T(f->dn,f->nFiles);
    for(int i=0;0&&i<chk->nSites;i++)
      fprintf(stderr,"pos[%d]\t%d\n",i,chk->nd[i][0].refPos+1);


    chk->regStart = f->regStart;
    chk->regStop = f->regStop;
    chk->refName = f->refName;
    //    fprintf(stderr,"regstart=%d regstop=%d\n",chk->regStart,chk->regStop);
    if(chk->refPos[0]>f->regStop){
      fprintf(stderr,"Problems with regions\n (chkFirst=%d lastReg=%d)\n",chk->refPos[0],f->regStop);
      fflush(stderr);
      exit(0);
    }

    //regstop and regstart are the usersupplied startstop region, strip ends if needed
    if(1&&(chk->nd[0][0].refPos<chk->regStart||chk->nd[chk->nSites-1][0].refPos>chk->regStop)){
      fprintf(stderr,"Stripping outof region parts of chunkID=%d nchunkNsites=%d\n",p->chunkNumber,chk->nSites);
      fprintf(stderr,"chk=(%d,%d)\tregs=(%d,%d)\n",chk->nd[0][0].refPos,chk->nd[chk->nSites-1][0].refPos,chk->regStart,chk->regStop);
      fflush(stderr);
      //exit(0);
      int at=0;
      for(int i=0;i<chk->nSites;i++){//can be written faster, daemon dragon nasty hack
	if(chk->nd[i][0].refPos<chk->regStart)//should cleanup
	  cleanUptNodeArray(chk->nd[i],chk->nSamples);
	//	  continue;
	else if(chk->nd[i][0].refPos<chk->regStop)
	  chk->nd[at++] = chk->nd[i];
	else{//should cleanup
	  //	  fprintf(stderr,"\ncleaning up i=%d\n",i);
	  cleanUptNodeArray(chk->nd[i],chk->nSamples);
	  //break;
	}
      }

      chk->nSites = at;
      //      fprintf(stderr,"nsties=%d\n",chk->nSites);
      
    }
    //now chk contains the merged data
    p->chk = chk;
    //now make the position vector
    p->sites =(loci*) populateLoci(chk);
    p->numSites=chk->nSites;
  
    p->nInd = chk->nSamples;  
    //fprintf(stderr,"after picking up\n");
  }
////////////////////////////////////////////////////////////////
  // keep all sites. p->keepSites keep track of the sites used in analysis and
  // printed to files
  //tsk keepInd is a vector of how many informative samples exists for a site
  /////////////////////////////////////////////////////////////
  p->keepSites=new int[p->numSites];
  p->results->freq->keepInd = new int[p->numSites];

  for(int s=0;s<p->numSites;s++){
    p->keepSites[s]=1;
    p->results->freq->keepInd[s]=p->nInd;
   
  }
  
  if(modSites)
    modify_keepsites(p);


  


  //  fprintf(stderr,"[%s] using nSites: %lu  (%s:%d)-(%s:%d)\n",__FUNCTION__,p->numSites,p->sites[0].chromo,p->sites[0].position,p->sites[p->numSites-1].chromo,p->sites[p->numSites-1].position);  

  //plugin ancestral fasta ,and reference fasta if -anc or -ref is supplied
  fastaObj->run(p);

  //now make GL using our the native method
  if(p->chk!=NULL){
    p->likes = gen_likesObj->call_likes(p->chk,p->ref);
    //catch SOAPsnp matrix recalibration
  }

  //make counts matrix if chunkT exists



  //countnucs, calculate phat, find major/minor(from nucs)
  cntObj->run(p);//phat only runs on keepsites==1,
  
  //make suyeons errorstuff
  errorObj->run(p);
  //make suyeons likestuff
  likesObj->run(p);
 


  //NOW we should have likes, let hcheck
  if(p->likes==NULL){
    return 0;
    fprintf(stderr,"Should have likes at this point in time and space (%s:%s():%d)\n",__FILE__,__FUNCTION__,__LINE__);
    fprintf(stderr,"supply -GL 0=soap -GL 1=sam -GL 2=gatk\n");
    exit(0);
    
  }
  //  fprintf(stderr,"JHAS LIKES\n");

  //keepsite moved up
  
  //this function also populates the keepInd vector which is a counter for how many individuals where we have reads
  //allocate the keepinf vector
 
  ////////////////////////////////////////////////////////////////
  //lets estimate the major minor. function defined in majorminor.cpp
  ////////////////////////////////////////////////////////////////

  majorminorObj->run(p);

  ////////////////////////////////////////////////////////////////
  // CALCULATE THE MAF
  ///////////////////////////////////////////////////////////////

  freqObj->run(p);

  //do HWE
  hweObj->run(p);


  assoObj->run(p);

 
  ///////////////////////////////////////////////////////
  // additional functions added like this 
  // weather they are run or not is defined in the class
  /////////////////////////////////////////////////////////
  covarObj->run(p);
  realSFSobj->run(p);

  return 1;
}


void deallocFunkyPars(funkyPars *p){
  //  fprintf(stderr,"[%s] cleaning funkypars\n",__FUNCTION__);  
  delete [] p->results->freq->keepInd;
  delete p->results->freq;
  delete p->results->asso;
  delete p->results;
  //  delete p->opt;
  delete p->covStruct;//TSK

  if(p->depth!=NULL){
    for(int i=0;i<p->numSites;i++)
      delete [] p->depth[i];
    delete [] p->depth;
  }

  //  fprintf(stderr,"[%s] p->sites=%p p->numistes=%lu\n",__FUNCTION__, p->sites,p->numSites);
  if(p->sites!=NULL){
    for(int i=0;i<p->numSites;i++)
      free(p->sites[i].chromo);
    delete [] p->sites;
  }
  if(1){
    if(p->anc!=NULL)
      delete []  p->anc;
    if(p->ref!=NULL)
      delete [] p->ref;
  }

  //cleanup extra stuff
  void cleanUpChunkyT(chunkyT *chk);//<- from mUpPile.h
  if(p->chk!=NULL){
    //    fprintf(stderr,"cleanuing up chunkyT\n");
    cleanUpChunkyT(p->chk);
    p->chk = NULL;
  }

  if(p->for_callback!=NULL){
    fcb *f = p->for_callback;
    for(int i=0;i<f->nFiles;i++)
      if(f->dn[i].l!=0)//this is nescesarry...
	delete [] f->dn[i].nds;
    delete [] f->dn;
    delete f;
    f=NULL;
  }
    
  delete p;

}
//plus one, plus 33
void printChunkyT(chunkyT *chk,double **liks,char *refs,FILE *fp){
  for(int s=0;s<chk->nSites;s++){
    fprintf(fp,"%s\t%d\t",chk->refName,chk->nd[s][0].refPos+1);
    if(refs!=NULL)
      fprintf(fp,"%c\t",intToRef[refs[s]]);
    for(int n=0;n<chk->nSamples;n++){
      tNode &nd = chk->nd[s][n];
      fprintf(fp,"%d\t",nd.l);
      for(int i=0;i<nd.l;i++)
	fprintf(fp,"%c",nd.seq[i]);
      fprintf(fp,"\t");
      for(int i=0;i<nd.l;i++)
	fprintf(fp,"%c",nd.qs[i]+33);
      fprintf(fp,"\t");
      for(int i=0;i<10;i++)
	fprintf(fp,"%f ",liks[s][n*10+i]);
    }
    fprintf(fp,"\n");
  }
}

//only one instance at a time is running this function
void printFunky(funkyPars *p){
 
  //fprintf(stderr,"printing chunk=%d\n",p->chunkNumber);

  if(p->killSig==0){//don't print the empty killSig chunk
    //    fprintf(stderr,"[%s] printing and cleaning: chunkNumber=%d\n",__FUNCTION__,p->chunkNumber);
    //printChunkyT(p->chk,p->likes,p->ref,fppp);// <- for debug

    //NB! the order is important!!!

    genotypesObj->printAndClean(p);//doesn cleanup anything,which is fine
    errorObj->printAndClean(p);
    likesObj->printAndClean(p);
    covarObj->printAndClean(p);
    assoObj->printAndClean(p);
    hweObj->printAndClean(p);
    freqObj->printAndClean(p);
    majorminorObj->printAndClean(p); //<-cleanup removed to downstairs
    realSFSobj->printAndClean(p);//<- doesnt do anything yet,result are being printet in the run;
    cntObj->printAndClean(p);//<- doesnt do anything yet,result are being printet in the run;

    //cleanup begin
    if(p->major!=NULL)
      delete[] p->major;
    if(p->minor!=NULL)
      delete[] p->minor;

  delete[] p->keepSites;
    
    for(int i=0;i<0&&i<p->numSites;p++)
      free(p->sites[i].chromo);
  }else
    pthread_mutex_unlock(&mUpPile_mutex);
  deallocFunkyPars(p);
  //  fprintf(stderr,"[%s] isEmpty=%d analysDone=%d\n",__FUNCTION__,printer.isEmpty(),analysisDone);


}













//////////////////////////////////////////////////////////////////////////
//   Stuff to remove from this file
//////////////////////////////////////////////////////////////////////////


// a,c,g,t,n
// A,C,G,T,N
// 0,1,2,3,4
int refToInt[256] = {
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//15
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//31
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//47
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//63
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//79
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//95
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//111
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//127
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//143
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//159
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//175
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//191
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//207
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//223
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//239
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4//255
};

char intToRef[5] = {'A','C','G','T','N'};



int fexists(const char* str){///@param str Filename given as a string.
  struct stat buffer ;
  return (stat(str, &buffer )==0 ); /// @return Function returns 1 if file exists.
}

size_t fsize(const char* fname){
  struct stat st ;
  stat(fname,&st);
  return st.st_size;
}

double* allocDoubleArray(int len){
  //  printf("allocing a double array:%d\n",len);
  double *ret= new double[len];
  for(int i=0;i<len;i++)
    ret[i]=0;
  return ret;
}

int* allocIntArray(int len){
  //  printf("allociPileupng a int array:%d\n",len);
  int *ret= new int[len];
  for(int i=0;i<len;i++)
    ret[i]=0;
  return ret;
}

std::vector <char *> dumpedFiles;//small hack for getting a nice vector of outputfiles
FILE *openFile(const char* a,const char* b){
  if(0)
    fprintf(stderr,"[%s] %s %s",__FUNCTION__,a,b);
  char *c = new char[strlen(a)+strlen(b)+1];
  strcpy(c,a);
  strncat(c,b,strlen(b));
  fprintf(stderr,"\t-> Dumping file: %s\n",c);
  if(0&&fexists(c)){//ANDERS DAEMON DRAGON HATES THIS
    fprintf(stderr,"File: %s exists will exist\n",c);
    fflush(stderr);
    exit(0);
  }
  dumpedFiles.push_back(strdup(c));
  FILE *fp = fopen(c,"w");
  delete [] c;
  return fp;
}
gzFile openFileGz(const char* a,const char* b,const char *mode){
  if(1)
    fprintf(stderr,"[%s] %s %s",__FUNCTION__,a,b);
  char *c = new char[strlen(a)+strlen(b)+1];
  strcpy(c,a);
  strncat(c,b,strlen(b));
  fprintf(stderr,"\t-> Dumping file: %s\n",c);
  if(0&&fexists(c)){//ANDERS DAEMON DRAGON HATES THIS
    fprintf(stderr,"File: %s exists will exist\n",c);
    fflush(stderr);
    exit(0);
  }
  dumpedFiles.push_back(strdup(c));
  gzFile fp = getGz(c,mode);
  delete [] c;
  return fp;
}


FILE *getFILE(const char*fname,const char* mode){
  int writeFile = 0;
  for(size_t i=0;i<strlen(mode);i++)
    if(mode[i]=='w')
      writeFile = 1;
  if(1&&writeFile&&fexists(fname)){//DRAGON
    fprintf(stderr,"\t-> File exists: %s exiting...\n",fname);
    exit(0);
  }
  FILE *fp;
  if(NULL==(fp=fopen(fname,mode))){
    fprintf(stderr,"\t->Error opening FILE handle for file:%s exiting\n",fname);
    exit(0);
  }
  return fp;
}

FILE *openFile(const char* a,const char* b,int i){
  char ary[100];//dragon
  snprintf(ary,100,"%s%d",b,i);
  return openFile(a,ary);  
}

void popBack(const char *a,const char*b,int to){
  char ary[100];//dragon
  snprintf(ary,100,"%s0-%d",b,to);
  dumpedFiles.push_back(strdup(ary));
}

void print_array(FILE *fp,double *ary,int len){
  for(int i=0;i<len-1;i++)
    fprintf(fp,"%f,",ary[i]);
  fprintf(fp,"%f\n",ary[len-1]);
}

void print_array(FILE *fp,int *ary,int len){
  for(int i=0;i<len-1;i++)
    fprintf(fp,"%d,",ary[i]);
  fprintf(fp,"%d\n",ary[len-1]);
}


gzFile getGz(const char*fname,const char* mode){
  int writeFile = 0;
  for(size_t i=0;i<strlen(mode);i++)
    if(mode[i]=='w')
      writeFile = 1;

  fprintf(stderr,"\t-> opening: %s\n",fname);
  if(0&&writeFile && fexists(fname)){//dragon
    fprintf(stderr,"\t-> File exists: %s exiting...\n",fname);
    exit(0);
  }
  gzFile fp=Z_NULL;
  if(NULL==(fp=gzopen(fname,mode))){
    fprintf(stderr,"\t-> Error opening gzFile handle for file:%s exiting\n",fname);
    exit(0);
  }
  return fp;
}

FILE *getFile(const char*fname,const char *mode){
  if(fexists(fname)){
    fprintf(stderr,"\t-> File exists: %s exiting...\n",fname);
    exit(0);
  }
  FILE *fp=NULL;
  if(NULL==(fp=fopen(fname,mode))){
    fprintf(stderr,"\t-> Error opening File handle for file:%s\n",fname);
    exit(0);
  }
  return fp;
}


/*
  initialize all pointers to zero
 */
funkyPars *allocFunkyPars(){
  //  fprintf(stderr,"[%s] making funkypars\n",__FUNCTION__);
  funkyPars *r = new funkyPars;

  r->sites = NULL;
  r->counts = NULL;
  r->likes = NULL;
  r->post = NULL;
  r->phat = NULL;
  //  r->errors = NULL;
  r->major = NULL;
  r->minor = NULL;
  r->depth = NULL;
  //  r->opt = NULL;
  r->results = NULL;
  r->ref = NULL;
  r->anc= NULL;
  r->keepSites =NULL;
  //allocate structures;
  //  r->opt = new funkyOpt;
  r->results = new funkyPrint;
  r->results->freq = new funkyPrintFreq;
  r->results->asso = new funkyPrintAsso;
  //set nullpointers
  r->results->asso->stat=NULL;
  r->results->asso->keepInd=NULL;
  r->results->asso->freq=NULL;
  r->results->asso->lrt_snp=NULL;
  r->results->freq->pml=NULL;
  r->results->freq->pmlun= NULL;
  r->results->freq->pEM=NULL;
  r->results->freq->pEMun=NULL;
  r->results->freq->keepInd=NULL;

  r->covStruct = new funkyCovar;//VALGRIND error here
  r->chk = NULL;
  r->for_callback = NULL;
  r->killSig =0;
  return r;
}
